@@ -1,7 +1,9 @@
 -- (c) The University of Glasgow 2006
 
-{-# LANGUAGE CPP, ScopedTypeVariables, ViewPatterns #-}
+
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE ViewPatterns #-}
 
 module GHC.Data.Graph.Directed (
         Graph, graphFromEdgedVerticesOrd, graphFromEdgedVerticesUniq,
@@ -10,7 +12,7 @@ module GHC.Data.Graph.Directed (
         stronglyConnCompG,
         topologicalSortG,
         verticesG, edgesG, hasVertexG,
-        reachableG, reachablesG, transposeG,
+        reachableG, reachablesG, transposeG, allReachable, outgoingG,
         emptyG,
 
         findCycle,
@@ -23,9 +25,7 @@ module GHC.Data.Graph.Directed (
 
         -- Simple way to classify edges
         EdgeType(..), classifyEdges
-    ) where
-
-#include "GhclibHsVersions.h"
+        ) where
 
 ------------------------------------------------------------------------------
 -- A version of the graph algorithms described in:
@@ -46,12 +46,13 @@ import GHC.Prelude
 
 import GHC.Utils.Misc ( minWith, count )
 import GHC.Utils.Outputable
+import GHC.Utils.Panic
 import GHC.Data.Maybe ( expectJust )
 
 -- std interfaces
 import Data.Maybe
 import Data.Array
-import Data.List hiding (transpose)
+import Data.List ( sort )
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 
@@ -60,6 +61,10 @@ import Data.Graph hiding (Graph, Edge, transposeG, reachable)
 import Data.Tree
 import GHC.Types.Unique
 import GHC.Types.Unique.FM
+import qualified Data.IntMap as IM
+import qualified Data.IntSet as IS
+import qualified Data.Map as M
+import qualified Data.Set as S
 
 {-
 ************************************************************************
@@ -105,7 +110,7 @@ data Node key payload = DigraphNode {
   }
 
 
-instance (Outputable a, Outputable b) => Outputable (Node  a b) where
+instance (Outputable a, Outputable b) => Outputable (Node a b) where
   ppr (DigraphNode a b c) = ppr (a, b, c)
 
 emptyGraph :: Graph a
@@ -358,6 +363,11 @@ reachableG graph from = map (gr_vertex_to_node graph) result
   where from_vertex = expectJust "reachableG" (gr_node_to_vertex graph from)
         result = {-# SCC "Digraph.reachable" #-} reachable (gr_int_graph graph) [from_vertex]
 
+outgoingG :: Graph node -> node -> [node]
+outgoingG graph from = map (gr_vertex_to_node graph) result
+  where from_vertex = expectJust "reachableG" (gr_node_to_vertex graph from)
+        result = gr_int_graph graph ! from_vertex
+
 -- | Given a list of roots return all reachable nodes.
 reachablesG :: Graph node -> [node] -> [node]
 reachablesG graph froms = map (gr_vertex_to_node graph) result
@@ -365,6 +375,15 @@ reachablesG graph froms = map (gr_vertex_to_node graph) result
                  reachable (gr_int_graph graph) vs
         vs = [ v | Just v <- map (gr_node_to_vertex graph) froms ]
 
+-- | Efficiently construct a map which maps each key to it's set of transitive
+-- dependencies.
+allReachable :: Ord key => Graph node -> (node -> key) -> M.Map key (S.Set key)
+allReachable (Graph g from _) conv =
+  M.fromList [(conv (from v), IS.foldr (\k vs -> conv (from k) `S.insert` vs) S.empty vs)
+             | (v, vs) <- IM.toList int_graph]
+  where
+    int_graph = reachableGraph g
+
 hasVertexG :: Graph node -> node -> Bool
 hasVertexG graph node = isJust $ gr_node_to_vertex graph node
 
@@ -434,6 +453,12 @@ preorderF ts         = concatMap flatten ts
 reachable    :: IntGraph -> [Vertex] -> [Vertex]
 reachable g vs = preorderF (dfs g vs)
 
+reachableGraph :: IntGraph -> IM.IntMap IS.IntSet
+reachableGraph g = res
+  where
+    do_one v = IS.unions (IS.fromList (g ! v) : mapMaybe (flip IM.lookup res) (g ! v))
+    res = IM.fromList [(v, do_one v) | v <- vertices g]
+
 {-
 ************************************************************************
 *                                                                      *
