@@ -14,12 +14,13 @@ therefore, is almost nothing but re-exporting.
 {-# LANGUAGE StandaloneDeriving #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE UndecidableInstances #-} -- Wrinkle in Note [Trees That Grow]
-                                      -- in module GHC.Hs.Extension
+                                      -- in module Language.Haskell.Syntax.Extension
 {-# LANGUAGE ConstraintKinds #-}
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE FlexibleInstances #-} -- For deriving instance Data
 
 module GHC.Hs (
+        module Language.Haskell.Syntax,
         module GHC.Hs.Binds,
         module GHC.Hs.Decls,
         module GHC.Hs.Expr,
@@ -30,9 +31,11 @@ module GHC.Hs (
         module GHC.Hs.Utils,
         module GHC.Hs.Doc,
         module GHC.Hs.Extension,
+        module GHC.Parser.Annotation,
         Fixity,
 
-        HsModule(..),
+        HsModule(..), AnnsModule(..),
+        HsParsedModule(..)
 ) where
 
 -- friends:
@@ -43,18 +46,21 @@ import GHC.Hs.Binds
 import GHC.Hs.Expr
 import GHC.Hs.ImpExp
 import GHC.Hs.Lit
+import Language.Haskell.Syntax
 import GHC.Hs.Extension
+import GHC.Parser.Annotation
 import GHC.Hs.Pat
 import GHC.Hs.Type
-import GHC.Types.Basic ( Fixity, WarningTxt )
 import GHC.Hs.Utils
 import GHC.Hs.Doc
 import GHC.Hs.Instances () -- For Data instances
 
 -- others:
 import GHC.Utils.Outputable
+import GHC.Types.Fixity         ( Fixity )
 import GHC.Types.SrcLoc
-import GHC.Unit.Module ( ModuleName )
+import GHC.Unit.Module          ( ModuleName )
+import GHC.Unit.Module.Warnings ( WarningTxt )
 
 -- libraries:
 import Data.Data hiding ( Fixity )
@@ -63,14 +69,24 @@ import Data.Data hiding ( Fixity )
 --
 -- All we actually declare here is the top-level structure for a module.
 data HsModule
-  = HsModule {
+  =  -- | 'GHC.Parser.Annotation.AnnKeywordId's
+     --
+     --  - 'GHC.Parser.Annotation.AnnModule','GHC.Parser.Annotation.AnnWhere'
+     --
+     --  - 'GHC.Parser.Annotation.AnnOpen','GHC.Parser.Annotation.AnnSemi',
+     --    'GHC.Parser.Annotation.AnnClose' for explicit braces and semi around
+     --    hsmodImports,hsmodDecls if this style is used.
+
+     -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
+    HsModule {
+      hsmodAnn :: EpAnn AnnsModule,
       hsmodLayout :: LayoutInfo,
         -- ^ Layout info for the module.
         -- For incomplete modules (e.g. the output of parseHeader), it is NoLayoutInfo.
-      hsmodName :: Maybe (Located ModuleName),
+      hsmodName :: Maybe (LocatedA ModuleName),
         -- ^ @Nothing@: \"module X where\" is omitted (in which case the next
         --     field is Nothing too)
-      hsmodExports :: Maybe (Located [LIE GhcPs]),
+      hsmodExports :: Maybe (LocatedL [LIE GhcPs]),
         -- ^ Export list
         --
         --  - @Nothing@: export list omitted, so export everything
@@ -83,51 +99,48 @@ data HsModule
         --  - 'GHC.Parser.Annotation.AnnKeywordId's : 'GHC.Parser.Annotation.AnnOpen'
         --                                   ,'GHC.Parser.Annotation.AnnClose'
 
-        -- For details on above see note [Api annotations] in GHC.Parser.Annotation
+        -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
       hsmodImports :: [LImportDecl GhcPs],
         -- ^ We snaffle interesting stuff out of the imported interfaces early
         -- on, adding that info to TyDecls/etc; so this list is often empty,
         -- downstream.
       hsmodDecls :: [LHsDecl GhcPs],
         -- ^ Type, class, value, and interface signature decls
-      hsmodDeprecMessage :: Maybe (Located WarningTxt),
+      hsmodDeprecMessage :: Maybe (LocatedP (WarningTxt GhcPs)),
         -- ^ reason\/explanation for warning/deprecation of this module
         --
         --  - 'GHC.Parser.Annotation.AnnKeywordId's : 'GHC.Parser.Annotation.AnnOpen'
         --                                   ,'GHC.Parser.Annotation.AnnClose'
         --
 
-        -- For details on above see note [Api annotations] in GHC.Parser.Annotation
-      hsmodHaddockModHeader :: Maybe LHsDocString
+        -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
+      hsmodHaddockModHeader :: Maybe (LHsDoc GhcPs)
         -- ^ Haddock module info and description, unparsed
         --
         --  - 'GHC.Parser.Annotation.AnnKeywordId's : 'GHC.Parser.Annotation.AnnOpen'
         --                                   ,'GHC.Parser.Annotation.AnnClose'
 
-        -- For details on above see note [Api annotations] in GHC.Parser.Annotation
+        -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
    }
-     -- ^ 'GHC.Parser.Annotation.AnnKeywordId's
-     --
-     --  - 'GHC.Parser.Annotation.AnnModule','GHC.Parser.Annotation.AnnWhere'
-     --
-     --  - 'GHC.Parser.Annotation.AnnOpen','GHC.Parser.Annotation.AnnSemi',
-     --    'GHC.Parser.Annotation.AnnClose' for explicit braces and semi around
-     --    hsmodImports,hsmodDecls if this style is used.
-
-     -- For details on above see note [Api annotations] in GHC.Parser.Annotation
 
 deriving instance Data HsModule
 
+data AnnsModule
+  = AnnsModule {
+    am_main :: [AddEpAnn],
+    am_decls :: AnnList
+    } deriving (Data, Eq)
+
 instance Outputable HsModule where
 
-    ppr (HsModule _ Nothing _ imports decls _ mbDoc)
-      = pp_mb mbDoc $$ pp_nonnull imports
-                    $$ pp_nonnull decls
+    ppr (HsModule _ _ Nothing _ imports decls _ mbDoc)
+      = pprMaybeWithDoc mbDoc $ pp_nonnull imports
+                             $$ pp_nonnull decls
 
-    ppr (HsModule _ (Just name) exports imports decls deprec mbDoc)
-      = vcat [
-            pp_mb mbDoc,
-            case exports of
+    ppr (HsModule _ _ (Just name) exports imports decls deprec mbDoc)
+      = pprMaybeWithDoc mbDoc $
+        vcat
+          [ case exports of
               Nothing -> pp_header (text "where")
               Just es -> vcat [
                            pp_header lparen,
@@ -144,10 +157,16 @@ instance Outputable HsModule where
 
         pp_modname = text "module" <+> ppr name
 
-pp_mb :: Outputable t => Maybe t -> SDoc
-pp_mb (Just x) = ppr x
-pp_mb Nothing  = empty
-
 pp_nonnull :: Outputable t => [t] -> SDoc
 pp_nonnull [] = empty
 pp_nonnull xs = vcat (map ppr xs)
+
+data HsParsedModule = HsParsedModule {
+    hpm_module    :: Located HsModule,
+    hpm_src_files :: [FilePath]
+       -- ^ extra source files (e.g. from #includes).  The lexer collects
+       -- these from '# <file> <line>' pragmas, which the C preprocessor
+       -- leaves behind.  These files and their timestamps are stored in
+       -- the .hi file, so that we can force recompilation if any of
+       -- them change (#3589)
+  }
