@@ -1,3 +1,11 @@
+{-# LANGUAGE DataKinds            #-}
+{-# LANGUAGE DeriveDataTypeable   #-}
+{-# LANGUAGE FlexibleContexts     #-}
+{-# LANGUAGE FlexibleInstances    #-}
+{-# LANGUAGE TypeApplications     #-}
+{-# LANGUAGE TypeFamilies         #-}
+{-# LANGUAGE UndecidableInstances #-} -- Wrinkle in Note [Trees That Grow]
+                                      -- in module Language.Haskell.Syntax.Extension
 {-
 (c) The University of Glasgow 2006
 (c) The GRASP/AQUA Project, Glasgow University, 1992-1998
@@ -6,28 +14,23 @@
 GHC.Hs.ImpExp: Abstract syntax: imports, exports, interfaces
 -}
 
-{-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE StandaloneDeriving #-}
-{-# LANGUAGE TypeFamilies #-}
-{-# LANGUAGE UndecidableInstances #-} -- Wrinkle in Note [Trees That Grow]
-                                      -- in module GHC.Hs.Extension
-
 module GHC.Hs.ImpExp where
 
 import GHC.Prelude
 
 import GHC.Unit.Module        ( ModuleName, IsBootInterface(..) )
-import GHC.Hs.Doc             ( HsDocString )
-import GHC.Types.Name.Occurrence ( HasOccName(..), isTcOcc, isSymOcc )
-import GHC.Types.Basic        ( SourceText(..), StringLiteral(..), pprWithSourceText )
-import GHC.Types.FieldLabel   ( FieldLbl(..) )
+import GHC.Hs.Doc
+import GHC.Types.SourceText   ( SourceText(..) )
+import GHC.Types.FieldLabel   ( FieldLabel )
 
 import GHC.Utils.Outputable
-import GHC.Data.FastString
+import GHC.Utils.Panic
 import GHC.Types.SrcLoc
+import Language.Haskell.Syntax.Extension
 import GHC.Hs.Extension
+import GHC.Parser.Annotation
+import GHC.Types.Name
+import GHC.Types.PkgQual
 
 import Data.Data
 import Data.Maybe
@@ -43,12 +46,13 @@ One per \tr{import} declaration in a module.
 -}
 
 -- | Located Import Declaration
-type LImportDecl pass = Located (ImportDecl pass)
+type LImportDecl pass = XRec pass (ImportDecl pass)
         -- ^ When in a list this may have
         --
         --  - 'GHC.Parser.Annotation.AnnKeywordId' : 'GHC.Parser.Annotation.AnnSemi'
 
-        -- For details on above see note [Api annotations] in GHC.Parser.Annotation
+        -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
+type instance Anno (ImportDecl (GhcPass p)) = SrcSpanAnnA
 
 -- | If/how an import is 'qualified'.
 data ImportDeclQualifiedStyle
@@ -60,12 +64,12 @@ data ImportDeclQualifiedStyle
 -- | Given two possible located 'qualified' tokens, compute a style
 -- (in a conforming Haskell program only one of the two can be not
 -- 'Nothing'). This is called from "GHC.Parser".
-importDeclQualifiedStyle :: Maybe (Located a)
-                         -> Maybe (Located a)
-                         -> ImportDeclQualifiedStyle
+importDeclQualifiedStyle :: Maybe EpaLocation
+                         -> Maybe EpaLocation
+                         -> (Maybe EpaLocation, ImportDeclQualifiedStyle)
 importDeclQualifiedStyle mPre mPost =
-  if isJust mPre then QualifiedPre
-  else if isJust mPost then QualifiedPost else NotQualified
+  if isJust mPre then (mPre, QualifiedPre)
+  else if isJust mPost then (mPost,QualifiedPost) else (Nothing, NotQualified)
 
 -- | Convenience function to answer the question if an import decl. is
 -- qualified.
@@ -80,15 +84,15 @@ data ImportDecl pass
   = ImportDecl {
       ideclExt       :: XCImportDecl pass,
       ideclSourceSrc :: SourceText,
-                                 -- Note [Pragma source text] in GHC.Types.Basic
-      ideclName      :: Located ModuleName, -- ^ Module name.
-      ideclPkgQual   :: Maybe StringLiteral,  -- ^ Package qualifier.
+                                 -- Note [Pragma source text] in GHC.Types.SourceText
+      ideclName      :: XRec pass ModuleName, -- ^ Module name.
+      ideclPkgQual   :: ImportDeclPkgQual pass,  -- ^ Package qualifier.
       ideclSource    :: IsBootInterface,      -- ^ IsBoot <=> {-\# SOURCE \#-} import
       ideclSafe      :: Bool,          -- ^ True => safe import
       ideclQualified :: ImportDeclQualifiedStyle, -- ^ If/how the import is qualified.
       ideclImplicit  :: Bool,          -- ^ True => implicit import (of Prelude)
-      ideclAs        :: Maybe (Located ModuleName),  -- ^ as Module
-      ideclHiding    :: Maybe (Bool, Located [LIE pass])
+      ideclAs        :: Maybe (XRec pass ModuleName),  -- ^ as Module
+      ideclHiding    :: Maybe (Bool, XRec pass [LIE pass])
                                        -- ^ (True => hiding, names)
     }
   | XImportDecl !(XXImportDecl pass)
@@ -107,17 +111,43 @@ data ImportDecl pass
      --    'GHC.Parser.Annotation.AnnClose' attached
      --     to location in ideclHiding
 
-     -- For details on above see note [Api annotations] in GHC.Parser.Annotation
+     -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
 
-type instance XCImportDecl  (GhcPass _) = NoExtField
-type instance XXImportDecl  (GhcPass _) = NoExtCon
+type family ImportDeclPkgQual pass
+type instance ImportDeclPkgQual GhcPs = RawPkgQual
+type instance ImportDeclPkgQual GhcRn = PkgQual
+type instance ImportDeclPkgQual GhcTc = PkgQual
 
-simpleImportDecl :: ModuleName -> ImportDecl (GhcPass p)
+type instance XCImportDecl  GhcPs = EpAnn EpAnnImportDecl
+type instance XCImportDecl  GhcRn = NoExtField
+type instance XCImportDecl  GhcTc = NoExtField
+
+type instance XXImportDecl  (GhcPass _) = DataConCantHappen
+
+type instance Anno ModuleName = SrcSpanAnnA
+type instance Anno [LocatedA (IE (GhcPass p))] = SrcSpanAnnL
+
+-- ---------------------------------------------------------------------
+
+-- API Annotations types
+
+data EpAnnImportDecl = EpAnnImportDecl
+  { importDeclAnnImport    :: EpaLocation
+  , importDeclAnnPragma    :: Maybe (EpaLocation, EpaLocation)
+  , importDeclAnnSafe      :: Maybe EpaLocation
+  , importDeclAnnQualified :: Maybe EpaLocation
+  , importDeclAnnPackage   :: Maybe EpaLocation
+  , importDeclAnnAs        :: Maybe EpaLocation
+  } deriving (Data)
+
+-- ---------------------------------------------------------------------
+
+simpleImportDecl :: ModuleName -> ImportDecl GhcPs
 simpleImportDecl mn = ImportDecl {
-      ideclExt       = noExtField,
+      ideclExt       = noAnn,
       ideclSourceSrc = NoSourceText,
-      ideclName      = noLoc mn,
-      ideclPkgQual   = Nothing,
+      ideclName      = noLocA mn,
+      ideclPkgQual   = NoRawPkgQual,
       ideclSource    = NotBoot,
       ideclSafe      = False,
       ideclImplicit  = False,
@@ -126,7 +156,9 @@ simpleImportDecl mn = ImportDecl {
       ideclHiding    = Nothing
     }
 
-instance OutputableBndrId p
+instance (OutputableBndrId p
+         , Outputable (Anno (IE (GhcPass p)))
+         , Outputable (ImportDeclPkgQual (GhcPass p)))
        => Outputable (ImportDecl (GhcPass p)) where
     ppr (ImportDecl { ideclSourceSrc = mSrcText, ideclName = mod'
                     , ideclPkgQual = pkg
@@ -134,15 +166,11 @@ instance OutputableBndrId p
                     , ideclQualified = qual, ideclImplicit = implicit
                     , ideclAs = as, ideclHiding = spec })
       = hang (hsep [text "import", ppr_imp from, pp_implicit implicit, pp_safe safe,
-                    pp_qual qual False, pp_pkg pkg, ppr mod', pp_qual qual True, pp_as as])
+                    pp_qual qual False, ppr pkg, ppr mod', pp_qual qual True, pp_as as])
              4 (pp_spec spec)
       where
         pp_implicit False = empty
-        pp_implicit True = ptext (sLit ("(implicit)"))
-
-        pp_pkg Nothing                    = empty
-        pp_pkg (Just (StringLiteral st p))
-          = pprWithSourceText st (doubleQuotes (ftext p))
+        pp_implicit True = text "(implicit)"
 
         pp_qual QualifiedPre False = text "qualified" -- Prepositive qualifier/prepositive position.
         pp_qual QualifiedPost True = text "qualified" -- Postpositive qualifier/postpositive position.
@@ -176,29 +204,32 @@ instance OutputableBndrId p
 ************************************************************************
 -}
 
--- | A name in an import or export specification which may have adornments. Used
--- primarily for accurate pretty printing of ParsedSource, and API Annotation
--- placement.
+-- | A name in an import or export specification which may have
+-- adornments. Used primarily for accurate pretty printing of
+-- ParsedSource, and API Annotation placement. The
+-- 'GHC.Parser.Annotation' is the location of the adornment in
+-- the original source.
 data IEWrappedName name
-  = IEName    (Located name)  -- ^ no extra
-  | IEPattern (Located name)  -- ^ pattern X
-  | IEType    (Located name)  -- ^ type (:+:)
+  = IEName                (LocatedN name)  -- ^ no extra
+  | IEPattern EpaLocation (LocatedN name)  -- ^ pattern X
+  | IEType    EpaLocation (LocatedN name)  -- ^ type (:+:)
   deriving (Eq,Data)
 
 -- | Located name with possible adornment
 -- - 'GHC.Parser.Annotation.AnnKeywordId's : 'GHC.Parser.Annotation.AnnType',
 --         'GHC.Parser.Annotation.AnnPattern'
-type LIEWrappedName name = Located (IEWrappedName name)
--- For details on above see note [Api annotations] in GHC.Parser.Annotation
+type LIEWrappedName name = LocatedA (IEWrappedName name)
+-- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
 
 
 -- | Located Import or Export
-type LIE pass = Located (IE pass)
+type LIE pass = XRec pass (IE pass)
         -- ^ When in a list this may have
         --
         --  - 'GHC.Parser.Annotation.AnnKeywordId' : 'GHC.Parser.Annotation.AnnComma'
 
-        -- For details on above see note [Api annotations] in GHC.Parser.Annotation
+        -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
+type instance Anno (IE (GhcPass p)) = SrcSpanAnnA
 
 -- | Imported or exported entity.
 data IE pass
@@ -212,7 +243,7 @@ data IE pass
         --  - 'GHC.Parser.Annotation.AnnKeywordId's : 'GHC.Parser.Annotation.AnnPattern',
         --             'GHC.Parser.Annotation.AnnType','GHC.Parser.Annotation.AnnVal'
 
-        -- For details on above see note [Api annotations] in GHC.Parser.Annotation
+        -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
         -- See Note [Located RdrNames] in GHC.Hs.Expr
   | IEThingAll  (XIEThingAll pass) (LIEWrappedName (IdP pass))
         -- ^ Imported or exported Thing with All imported or exported
@@ -223,14 +254,13 @@ data IE pass
         --       'GHC.Parser.Annotation.AnnDotdot','GHC.Parser.Annotation.AnnClose',
         --                                 'GHC.Parser.Annotation.AnnType'
 
-        -- For details on above see note [Api annotations] in GHC.Parser.Annotation
+        -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
         -- See Note [Located RdrNames] in GHC.Hs.Expr
 
   | IEThingWith (XIEThingWith pass)
                 (LIEWrappedName (IdP pass))
                 IEWildcard
                 [LIEWrappedName (IdP pass)]
-                [Located (FieldLbl (IdP pass))]
         -- ^ Imported or exported Thing With given imported or exported
         --
         -- The thing is a Class/Type and the imported or exported things are
@@ -240,29 +270,42 @@ data IE pass
         --                                   'GHC.Parser.Annotation.AnnComma',
         --                                   'GHC.Parser.Annotation.AnnType'
 
-        -- For details on above see note [Api annotations] in GHC.Parser.Annotation
-  | IEModuleContents  (XIEModuleContents pass) (Located ModuleName)
+        -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
+  | IEModuleContents  (XIEModuleContents pass) (XRec pass ModuleName)
         -- ^ Imported or exported module contents
         --
         -- (Export Only)
         --
         -- - 'GHC.Parser.Annotation.AnnKeywordId's : 'GHC.Parser.Annotation.AnnModule'
 
-        -- For details on above see note [Api annotations] in GHC.Parser.Annotation
-  | IEGroup             (XIEGroup pass) Int HsDocString -- ^ Doc section heading
-  | IEDoc               (XIEDoc pass) HsDocString       -- ^ Some documentation
+        -- For details on above see Note [exact print annotations] in GHC.Parser.Annotation
+  | IEGroup             (XIEGroup pass) Int (LHsDoc pass) -- ^ Doc section heading
+  | IEDoc               (XIEDoc pass) (LHsDoc pass)       -- ^ Some documentation
   | IEDocNamed          (XIEDocNamed pass) String    -- ^ Reference to named doc
   | XIE !(XXIE pass)
 
-type instance XIEVar             (GhcPass _) = NoExtField
-type instance XIEThingAbs        (GhcPass _) = NoExtField
-type instance XIEThingAll        (GhcPass _) = NoExtField
-type instance XIEThingWith       (GhcPass _) = NoExtField
-type instance XIEModuleContents  (GhcPass _) = NoExtField
+type instance XIEVar             GhcPs = NoExtField
+type instance XIEVar             GhcRn = NoExtField
+type instance XIEVar             GhcTc = NoExtField
+
+type instance XIEThingAbs        (GhcPass _) = EpAnn [AddEpAnn]
+type instance XIEThingAll        (GhcPass _) = EpAnn [AddEpAnn]
+
+-- See Note [IEThingWith]
+type instance XIEThingWith       (GhcPass 'Parsed)      = EpAnn [AddEpAnn]
+type instance XIEThingWith       (GhcPass 'Renamed)     = [Located FieldLabel]
+type instance XIEThingWith       (GhcPass 'Typechecked) = NoExtField
+
+type instance XIEModuleContents  GhcPs = EpAnn [AddEpAnn]
+type instance XIEModuleContents  GhcRn = NoExtField
+type instance XIEModuleContents  GhcTc = NoExtField
+
 type instance XIEGroup           (GhcPass _) = NoExtField
 type instance XIEDoc             (GhcPass _) = NoExtField
 type instance XIEDocNamed        (GhcPass _) = NoExtField
-type instance XXIE               (GhcPass _) = NoExtCon
+type instance XXIE               (GhcPass _) = DataConCantHappen
+
+type instance Anno (LocatedA (IE (GhcPass p))) = SrcSpanAnnA
 
 -- | Imported or Exported Wildcard
 data IEWildcard = NoIEWildcard | IEWildcard Int deriving (Eq, Data)
@@ -270,53 +313,67 @@ data IEWildcard = NoIEWildcard | IEWildcard Int deriving (Eq, Data)
 {-
 Note [IEThingWith]
 ~~~~~~~~~~~~~~~~~~
-
 A definition like
 
+    {-# LANGUAGE DuplicateRecordFields #-}
     module M ( T(MkT, x) ) where
       data T = MkT { x :: Int }
 
-gives rise to
+gives rise to this in the output of the parser:
 
-    IEThingWith T [MkT] [FieldLabel "x" False x)]           (without DuplicateRecordFields)
-    IEThingWith T [MkT] [FieldLabel "x" True $sel:x:MkT)]   (with    DuplicateRecordFields)
+    IEThingWith NoExtField T [MkT, x] NoIEWildcard
+
+But in the renamer we need to attach the correct field label,
+because the selector Name is mangled (see Note [FieldLabel] in
+GHC.Types.FieldLabel).  Hence we change this to:
+
+    IEThingWith [FieldLabel "x" True $sel:x:MkT)] T [MkT] NoIEWildcard
+
+using the TTG extension field to store the list of fields in renamed syntax
+only.  (Record fields always appear in this list, regardless of whether
+DuplicateRecordFields was in use at the definition site or not.)
 
 See Note [Representing fields in AvailInfo] in GHC.Types.Avail for more details.
 -}
 
 ieName :: IE (GhcPass p) -> IdP (GhcPass p)
-ieName (IEVar _ (L _ n))              = ieWrappedName n
-ieName (IEThingAbs  _ (L _ n))        = ieWrappedName n
-ieName (IEThingWith _ (L _ n) _ _ _)  = ieWrappedName n
-ieName (IEThingAll  _ (L _ n))        = ieWrappedName n
+ieName (IEVar _ (L _ n))            = ieWrappedName n
+ieName (IEThingAbs  _ (L _ n))      = ieWrappedName n
+ieName (IEThingWith _ (L _ n) _ _)  = ieWrappedName n
+ieName (IEThingAll  _ (L _ n))      = ieWrappedName n
 ieName _ = panic "ieName failed pattern match!"
 
 ieNames :: IE (GhcPass p) -> [IdP (GhcPass p)]
-ieNames (IEVar       _ (L _ n)   )     = [ieWrappedName n]
-ieNames (IEThingAbs  _ (L _ n)   )     = [ieWrappedName n]
-ieNames (IEThingAll  _ (L _ n)   )     = [ieWrappedName n]
-ieNames (IEThingWith _ (L _ n) _ ns _) = ieWrappedName n
-                                       : map (ieWrappedName . unLoc) ns
+ieNames (IEVar       _ (L _ n)   )   = [ieWrappedName n]
+ieNames (IEThingAbs  _ (L _ n)   )   = [ieWrappedName n]
+ieNames (IEThingAll  _ (L _ n)   )   = [ieWrappedName n]
+ieNames (IEThingWith _ (L _ n) _ ns) = ieWrappedName n
+                                     : map (ieWrappedName . unLoc) ns
+-- NB the above case does not include names of field selectors
 ieNames (IEModuleContents {})     = []
 ieNames (IEGroup          {})     = []
 ieNames (IEDoc            {})     = []
 ieNames (IEDocNamed       {})     = []
 
+ieWrappedLName :: IEWrappedName name -> LocatedN name
+ieWrappedLName (IEName      ln) = ln
+ieWrappedLName (IEPattern _ ln) = ln
+ieWrappedLName (IEType    _ ln) = ln
+
 ieWrappedName :: IEWrappedName name -> name
-ieWrappedName (IEName    (L _ n)) = n
-ieWrappedName (IEPattern (L _ n)) = n
-ieWrappedName (IEType    (L _ n)) = n
+ieWrappedName = unLoc . ieWrappedLName
+
 
 lieWrappedName :: LIEWrappedName name -> name
 lieWrappedName (L _ n) = ieWrappedName n
 
-ieLWrappedName :: LIEWrappedName name -> Located name
-ieLWrappedName (L l n) = L l (ieWrappedName n)
+ieLWrappedName :: LIEWrappedName name -> LocatedN name
+ieLWrappedName (L _ n) = ieWrappedLName n
 
 replaceWrappedName :: IEWrappedName name1 -> name2 -> IEWrappedName name2
-replaceWrappedName (IEName    (L l _)) n = IEName    (L l n)
-replaceWrappedName (IEPattern (L l _)) n = IEPattern (L l n)
-replaceWrappedName (IEType    (L l _)) n = IEType    (L l n)
+replaceWrappedName (IEName      (L l _)) n = IEName      (L l n)
+replaceWrappedName (IEPattern r (L l _)) n = IEPattern r (L l n)
+replaceWrappedName (IEType    r (L l _)) n = IEType    r (L l n)
 
 replaceLWrappedName :: LIEWrappedName name1 -> name2 -> LIEWrappedName name2
 replaceLWrappedName (L l n) n' = L l (replaceWrappedName n n')
@@ -325,10 +382,9 @@ instance OutputableBndrId p => Outputable (IE (GhcPass p)) where
     ppr (IEVar       _     var) = ppr (unLoc var)
     ppr (IEThingAbs  _   thing) = ppr (unLoc thing)
     ppr (IEThingAll  _   thing) = hcat [ppr (unLoc thing), text "(..)"]
-    ppr (IEThingWith _ thing wc withs flds)
+    ppr (IEThingWith flds thing wc withs)
         = ppr (unLoc thing) <> parens (fsep (punctuate comma
-                                              (ppWiths ++
-                                              map (ppr . flLabel . unLoc) flds)))
+                                              (ppWiths ++ ppFields) ))
       where
         ppWiths =
           case wc of
@@ -337,6 +393,10 @@ instance OutputableBndrId p => Outputable (IE (GhcPass p)) where
               IEWildcard pos ->
                 let (bs, as) = splitAt pos (map (ppr . unLoc) withs)
                 in bs ++ [text ".."] ++ as
+        ppFields =
+          case ghcPass @p of
+            GhcRn -> map ppr flds
+            _     -> []
     ppr (IEModuleContents _ mod')
         = text "module" <+> ppr mod'
     ppr (IEGroup _ n _)           = text ("<IEGroup: " ++ show n ++ ">")
@@ -352,9 +412,9 @@ instance (OutputableBndr name) => OutputableBndr (IEWrappedName name) where
   pprInfixOcc  w = pprInfixOcc  (ieWrappedName w)
 
 instance (OutputableBndr name) => Outputable (IEWrappedName name) where
-  ppr (IEName    n) = pprPrefixOcc (unLoc n)
-  ppr (IEPattern n) = text "pattern" <+> pprPrefixOcc (unLoc n)
-  ppr (IEType    n) = text "type"    <+> pprPrefixOcc (unLoc n)
+  ppr (IEName      n) = pprPrefixOcc (unLoc n)
+  ppr (IEPattern _ n) = text "pattern" <+> pprPrefixOcc (unLoc n)
+  ppr (IEType    _ n) = text "type"    <+> pprPrefixOcc (unLoc n)
 
 pprImpExp :: (HasOccName name, OutputableBndr name) => name -> SDoc
 pprImpExp name = type_pref <+> pprPrefixOcc name
